import json

import httpx
import pytest
from respx import MockRouter
from tornado.httpclient import HTTPRequest
from tornado.httputil import HTTPHeaders

from aext_shared.backend_proxy import ProxyResponse, backend_proxy
from aext_shared.errors import BackendError, BadRequest, UnknownError


@pytest.fixture()
def incoming():
    return HTTPRequest(
        "https://example.com",
        headers={"User-Agent": "firefox", "Api-Version": "2023-02-25"},
    )


async def test_allows_user_agent_override(respx_mock: MockRouter, incoming):
    respx_mock.get()
    headers = {"User-Agent": "abc"}
    assert await backend_proxy(incoming, "https://anaconda.cloud/api", headers=headers) == ProxyResponse(
        remote_status_code=200, remote_data=None
    )
    # Ensure the original request headers weren't modified
    assert headers["User-Agent"] == "abc"
    proxy_request = respx_mock.calls.last.request
    assert proxy_request.headers["User-Agent"] == "abc"


async def test_allows_api_version_override_with_http_headers(respx_mock: MockRouter, incoming):
    respx_mock.get()
    headers = HTTPHeaders({"Api-Version": "abc"})
    assert await backend_proxy(incoming, "https://anaconda.cloud/api", headers=headers) == ProxyResponse(
        remote_status_code=200, remote_data=None
    )
    proxy_request = respx_mock.calls.last.request
    assert proxy_request.headers["Api-Version"] == "abc"


async def test_connection_error(respx_mock: MockRouter, incoming):
    respx_mock.get().mock(side_effect=httpx.ReadTimeout("Unable to read within timeout"))
    with pytest.raises(UnknownError, match="proxy_connection_error"):
        await backend_proxy(incoming, "https://anaconda.cloud/api")


async def test_valid_no_content(respx_mock: MockRouter, incoming):
    respx_mock.get().respond(200)
    headers = {}
    assert await backend_proxy(incoming, "https://anaconda.cloud/api", headers=headers) == ProxyResponse(
        remote_status_code=200, remote_data=None
    )

    # Ensure the original request headers weren't modified
    assert headers == {}
    # But make sure the user agent was passed through
    proxy_request = respx_mock.calls.last.request
    assert proxy_request.headers["User-Agent"] == "firefox"


async def test_400_no_content(respx_mock: MockRouter, incoming):
    respx_mock.get().respond(status_code=400)
    with pytest.raises(BackendError) as exc_info:
        await backend_proxy(incoming, "https://anaconda.cloud/api")
    assert isinstance(exc_info.value, BackendError)
    assert exc_info.value.status_code == 200
    assert exc_info.value.data == {
        "remote_status_code": 400,
        "remote_data": None,
        "reason": "proxy_bad_response",
    }


async def test_valid_with_json(respx_mock: MockRouter, incoming):
    respx_mock.get().respond(json={"hello": "world"})
    assert await backend_proxy(incoming, "https://anaconda.cloud/api") == ProxyResponse(
        remote_status_code=200, remote_data={"hello": "world"}
    )


async def test_send_json_to_proxy(respx_mock: MockRouter, incoming):
    respx_mock.post().respond(json={"command": "pong"})
    assert await backend_proxy(
        incoming, "https://anaconda.cloud/api", json={"command": "ping"}, method="POST"
    ) == ProxyResponse(remote_status_code=200, remote_data={"command": "pong"})
    proxy_request = respx_mock.calls.last.request
    assert proxy_request.headers["Content-Type"] == "application/json"
    assert proxy_request.content == json.dumps({"command": "ping"}).encode("utf-8")


async def test_500_with_json(respx_mock: MockRouter, incoming):
    respx_mock.get().respond(
        status_code=500,
        json={"reason": "nope"},
    )
    with pytest.raises(BackendError) as exc_info:
        await backend_proxy(incoming, "https://anaconda.cloud/api")
    assert isinstance(exc_info.value, BackendError)
    assert exc_info.value.status_code == 200
    assert exc_info.value.data == {
        "remote_status_code": 500,
        "remote_data": {"reason": "nope"},
        "reason": "proxy_bad_response",
    }


async def test_valid_with_bad_json_response(respx_mock: MockRouter, incoming):
    respx_mock.get().respond(text="[not_json}")
    with pytest.raises(BadRequest, match="proxy_response_not_json"):
        await backend_proxy(incoming, "https://anaconda.cloud/api")


async def test_502_with_bad_json_response(respx_mock: MockRouter, incoming):
    respx_mock.get().respond(
        status_code=502,
        text="[not_json}",
    )
    with pytest.raises(BadRequest, match="proxy_response_not_json"):
        await backend_proxy(incoming, "https://anaconda.cloud/api")



import json
from http.cookies import SimpleCookie
from typing import Dict, Optional, Type
from unittest.mock import patch

import jwt
import respx
import tornado.web
from anaconda_cloud_auth.exceptions import TokenNotFoundError
from anaconda_cloud_auth.token import TokenInfo
from httpx import Request, Response
from jupyter_server.auth.identity import IdentityProvider
from tornado.httpclient import HTTPResponse
from tornado.testing import AsyncHTTPTestCase

import aext_shared.handler
from aext_shared.handler import BackendError, BackendHandler, create_rules


class GoodHandler(BackendHandler):
    @tornado.web.authenticated
    async def get(self):
        self.finish({"hello": "world"})


class AnacondaCloudHandler(BackendHandler):
    @tornado.web.authenticated
    async def get(self):
        response = await self.anaconda_cloud_proxy("account", method="GET", headers={"cloud_header": "cloud_value"})
        self.finish(response)


class AnacondaPublicHandler(BackendHandler):
    @tornado.web.authenticated
    async def get(self):
        response = await self.anaconda_cloud_proxy("public_data", method="GET", auth_optional=True)
        self.finish(response)


class RaisesHttpErrorHandler(BackendHandler):
    @tornado.web.authenticated
    async def get(self):
        status_code = int(self.request.headers.get("Response-Code", "500"))
        raise tornado.web.HTTPError(status_code=status_code, log_message="uh oh")


class RaisesBackendErrorHandler(BackendHandler):
    @tornado.web.authenticated
    async def get(self):
        status_code = int(self.request.headers.get("Response-Code", "500"))
        data = json.loads(self.request.headers.get("Response-Payload", "{}"))
        log_message = self.request.headers.get("Log-Message", None)
        raise BackendError(status_code, data, log_message)


class TestBackendHandler(AsyncHTTPTestCase):
    def setUp(self):
        # Mocked anaconda.cloud API httpx responses
        aext_shared.handler.extra_anaconda_headers = {
            "testbackendhandler": "extra_header",
            "testbackendhandler2": "extra_header2",
        }
        self.account_response = Response(
            status_code=200,
            json={
                "profile": {"is_confirmed": True, "is_disabled": False},
                "subscriptions": [
                    {"id": "sub_1"},
                    {"id": "sub_2"},
                ],
                "user": {"id": "user_1", "email": "snek@anaconda.com"},
            },
        )
        self.token_response = Response(
            status_code=200,
            json={
                "access_token": "deadbeef",
                "expires_in": 2000,
            },
        )
        self.mocked_api = respx.mock(base_url="https://anaconda.cloud/api/", assert_all_called=False)
        self.mocked_api.get("account").mock(side_effect=self.account_side_effect)
        self.mocked_api.post("iam/token").mock(side_effect=self.token_side_effect)
        self.mocked_api.get("public_data").mock(return_value=self.account_response)
        self.mocked_api.start()
        super().setUp()

    def tearDown(self) -> None:
        aext_shared.handler.extra_anaconda_headers = {}
        self.mocked_api.stop()
        return super().tearDown()

    def account_side_effect(self, request: Request) -> Response:
        if request.headers.get("Authorization") == "Bearer deadbeef":
            return self.account_response
        return Response(status_code=401)

    def token_side_effect(self, request: Request) -> Response:
        payload = json.loads(request.content)
        if payload.get("refresh_token") == "xyz":
            return self.token_response
        return Response(status_code=400, json="oh no")

    def get_app(self):
        handlers: Dict[str, Type[BackendHandler]] = {
            "good": GoodHandler,
            "anaconda": AnacondaCloudHandler,
            "anaconda_public": AnacondaPublicHandler,
            "raise_http_error": RaisesHttpErrorHandler,
            "raise_backend_error": RaisesBackendErrorHandler,
        }
        return tornado.web.Application(
            create_rules("/", "", handlers),
            cookie_secret="abc",
            identity_provider=IdentityProvider(token="abc"),
        )

    def test_good_handler(self):
        headers = {"Authorization": "token abc"}
        response = self.fetch("/good", headers=headers)
        assert response.code == 200
        assert json.loads(response.body) == {"hello": "world"}

    def test_anaconda_handler_missing_refresh_token(self):
        headers = {"Authorization": "token abc"}

        @classmethod
        def mock_load(cls, domain):
            raise TokenNotFoundError

        with patch.object(TokenInfo, "load", mock_load):
            response = self.fetch("/anaconda", headers=headers)
        assert response.code == 403
        assert json.loads(response.body) == {"reason": "missing refresh_token"}

    def test_anaconda_handler_valid_access_token(self):
        headers = {
            "Authorization": "token abc",
            "Cookie": _to_cookie(access_token="deadbeef", refresh_token="xyz"),
        }
        response = self.fetch("/anaconda", headers=headers)
        assert response.code == 200
        assert json.loads(response.body) == {
            "remote_data": self.account_response.json(),
            "remote_status_code": 200,
        }

    def test_anaconda_handler_valid_access_token_bad_refresh_token(self):
        headers = {
            "Authorization": "token abc",
            "Cookie": _to_cookie(access_token="deadbeef", refresh_token="bad_refresh_token"),
        }
        response = self.fetch("/anaconda", headers=headers)
        assert response.code == 200
        assert json.loads(response.body) == {
            "remote_data": self.account_response.json(),
            "remote_status_code": 200,
        }

    def test_anaconda_handler_valid_refresh_token_only(self):
        headers = {
            "Authorization": "token abc",
            "Cookie": _to_cookie(refresh_token="xyz"),
        }
        response = self.fetch("/anaconda", headers=headers)
        assert response.code == 200
        assert get_access_cookie_from_response(response) == "deadbeef"
        assert json.loads(response.body) == {
            "remote_data": self.account_response.json(),
            "remote_status_code": 200,
        }
        for call in self.mocked_api.calls:
            assert call.request.headers.get("testbackendhandler") == "extra_header"
            assert call.request.headers.get("testbackendhandler2") == "extra_header2"

    def test_anaconda_handler_invalid_refresh_token_only(self):
        headers = {
            "Authorization": "token abc",
            "Cookie": _to_cookie(refresh_token="invalid"),
        }
        response = self.fetch("/anaconda", headers=headers)
        assert response.code == 403
        assert json.loads(response.body) == {
            "remote_status_code": 400,
            "reason": "proxy_bad_response",
            "remote_data": "oh no",
        }

    def test_anaconda_handler_expired_access_token(self):
        headers = {
            "Authorization": "token abc",
            "Cookie": _to_cookie(access_token="expired", refresh_token="xyz"),
        }
        response = self.fetch("/anaconda", headers=headers)
        assert response.code == 200
        assert json.loads(response.body) == {
            "remote_data": self.account_response.json(),
            "remote_status_code": 200,
        }
        assert get_access_cookie_from_response(response) == "deadbeef"
        assert self.mocked_api.calls.last.request.headers.get("cloud_header") == "cloud_value"

    def test_anaconda_handler_api_key(self):
        """We can also use an API key for authentication."""
        headers = {
            "Authorization": "token abc",  # Doesn't matter what it is
        }

        @classmethod
        def mock_load(cls, domain):
            return TokenInfo(api_key="deadbeef", domain=domain)

        with patch.object(TokenInfo, "load", mock_load):
            response = self.fetch("/anaconda", headers=headers)
        assert response.code == 200
        assert json.loads(response.body) == {
            "remote_data": self.account_response.json(),
            "remote_status_code": 200,
        }

    def test_anaconda_handler_expired_access_token_and_bad_refresh_token(self):
        headers = {
            "Authorization": "token abc",
            "Cookie": _to_cookie(access_token="expired", refresh_token="bad_refresh_token"),
        }
        response = self.fetch("/anaconda", headers=headers)
        assert response.code == 403
        assert json.loads(response.body) == {
            "remote_status_code": 400,
            "reason": "proxy_bad_response",
            "remote_data": "oh no",
        }
        assert get_access_cookie_from_response(response) is None

    def test_anaconda_handler_return_400(self):
        headers = {
            "Authorization": "token abc",
            "Cookie": _to_cookie(access_token="deadbeef", refresh_token="xyz"),
        }
        self.account_response = Response(
            status_code=400,
            json={"message": "some downstream error"},
        )
        response = self.fetch("/anaconda", headers=headers)
        assert response.code == 200
        assert json.loads(response.body) == {
            "reason": "proxy_bad_response",
            "remote_status_code": 400,
            "remote_data": self.account_response.json(),
        }
        assert self.mocked_api.calls.last.request.headers.get("cloud_header") == "cloud_value"

    def test_anaconda_handler_expired_access_token_return_400(self):
        headers = {
            "Authorization": "token abc",
            "Cookie": _to_cookie(access_token="expired", refresh_token="xyz"),
        }
        self.account_response = Response(
            status_code=400,
            json={"message": "some downstream error"},
        )
        response = self.fetch("/anaconda", headers=headers)
        assert response.code == 200
        assert json.loads(response.body) == {
            "reason": "proxy_bad_response",
            "remote_status_code": 400,
            "remote_data": self.account_response.json(),
        }

    def test_anaconda_handler_only_refresh_token_downstream_returns_403(self):
        headers = {
            "Authorization": "token abc",
            "Cookie": _to_cookie(refresh_token="xyz"),
        }
        self.account_response = Response(
            status_code=403,
            json={"message": "downstream says unauthorized"},
        )
        response = self.fetch("/anaconda", headers=headers)
        assert response.code == 200
        assert json.loads(response.body) == {
            "reason": "proxy_bad_response",
            "remote_status_code": 403,
            "remote_data": self.account_response.json(),
        }
        assert get_access_cookie_from_response(response) == "deadbeef"

    def test_raise_http_400(self):
        headers = {"Authorization": "token abc", "Response-Code": "400"}
        response = self.fetch("/raise_http_error", headers=headers)
        assert response.code == 400
        assert json.loads(response.body) == {"reason": "unknown"}

    def test_raise_http_502(self):
        headers = {"Authorization": "token abc", "Response-Code": "502"}
        response = self.fetch("/raise_http_error", headers=headers)
        assert response.code == 502
        assert json.loads(response.body) == {"reason": "unknown"}

    def test_raise_unknown_backend_400(self):
        headers = {
            "Authorization": "token abc",
            "Response-Code": "400",
            "Response-Payload": json.dumps({"other": "data"}),
        }
        response = self.fetch("/raise_backend_error", headers=headers)
        assert response.code == 400
        assert json.loads(response.body) == {"reason": "unknown", "other": "data"}

    def test_raise_backend_500_with_reason(self):
        headers = {
            "Authorization": "token abc",
            "Response-Code": "500",
            "Response-Payload": json.dumps({"reason": "nope", "other": "data"}),
        }
        response = self.fetch("/raise_backend_error", headers=headers)
        assert response.code == 500
        assert json.loads(response.body) == {"reason": "nope", "other": "data"}

    def test_auth_optional_200_without_auth(self):
        headers = {
            "Authorization": "token abc",
        }
        response = self.fetch("/anaconda_public", headers=headers)
        assert response.code == 200
        assert json.loads(response.body) == {
            "remote_data": self.account_response.json(),
            "remote_status_code": 200,
        }

    def test_auth_optional_200_with_only_refresh_token(self):
        headers = {
            "Authorization": "token abc",
            "Cookie": _to_cookie(refresh_token="xyz"),
        }
        response = self.fetch("/anaconda_public", headers=headers)
        assert response.code == 200
        assert json.loads(response.body) == {
            "remote_data": self.account_response.json(),
            "remote_status_code": 200,
        }
        assert get_access_cookie_from_response(response) == "deadbeef"


def test_backend_error_unknown_reason():
    err = BackendError(status_code=400, data={})
    assert str(err) == "BackendError 400: unknown"


def test_backend_error_with_reason():
    err = BackendError(status_code=400, data={"reason": "nope"})
    assert str(err) == "BackendError 400: nope"


def test_backend_error_custom_message():
    err = BackendError(status_code=400, data={"reason": "nope"}, log_message="custom message")
    assert str(err) == "custom message"


def test_is_token_expired():
    assert (
        BackendHandler._is_token_expired(
            "eyJhbGciOiJSUzI1NiIsImtpZCI6ImRlZmF1bHQiLCJ0eXAiOiJKV1QifQ.eyJleHAiOjE3MDcyOTg3NTYsInN1YiI6ImU3NjFkYTU4LTczMmUtNDk0My1hZjQwLTVkNGNhYmE0MWJmZSJ9.etZ_HMmCBbcTlUT4CEGmd1H0ov4L0To2-ldNltG1whM2KwG9U9wXgTi61kkd4Lv9eTChzELCTDuAVSFEA0eCwrC4txCO_Y4yUU6CuAmbR7s0tbtB599s7u44m_sj8IxrzF_IoyTgfU-WAgNhEXnDNWo33X8ZR174EteN1Ol_FvFQo-mhKDLRyIJkewFFpxepsfCLz3wYnpBAq8XBIzmtYBSnkByKH8sPE-FAGM2Z8HLXY7j8vGY1cORm5dM-Anlis_ffy-3jnoaaFzOXeCqlaR1xp_2ShDgx01E76h9wDIhHhCbyhaRtmnn7wS3mkKuqMNLN0GO3QgtmtbH4BH67aw"
        )
        is True
    )


def get_access_cookie_from_response(response: HTTPResponse) -> Optional[str]:
    headers = response.headers.get_list("Set-Cookie")
    access_token = None
    for header in headers:
        cookie = SimpleCookie()
        cookie.load(header)
        # Make sure we don't try to set the access_token twice
        assert (access_token is None) or (cookie.get("access_token") is None)
        val = cookie.get("access_token")
        if val is not None:
            access_token = val.value
    return access_token


def _to_cookie(access_token: Optional[str] = None, refresh_token: Optional[str] = None):
    vals = {"access_token": access_token, "refresh_token": refresh_token}
    return "; ".join([f"{key}={value}" for key, value in vals.items() if value is not None])



import pytest

from aext_shared.errors import BackendError, BadRequest, UnknownError


@pytest.mark.parametrize(
    ["error", "status_code", "message"],
    [
        [BackendError(400, {"other": "data"}), 400, "BackendError 400: unknown"],
        [BackendError(500, {"reason": "nope"}), 500, "BackendError 500: nope"],
        [UnknownError({"reason": "nope"}), 500, "BackendError 500: nope"],
        [UnknownError({}), 500, "BackendError 500: unknown"],
        [BadRequest({}), 400, "BackendError 400: bad_request"],
        [BadRequest({"reason": "nope"}), 400, "BackendError 400: nope"],
    ],
)
def test_errors(error: BackendError, status_code, message):
    assert error.status_code == status_code
    assert str(error) == message



from typing import Any, Dict, Optional


class BackendError(Exception):
    status_code: int
    data: Dict[str, Any]

    def __init__(self, status_code: int, data: Dict[str, Any], log_message: Optional[str] = None):
        if "reason" not in data:
            data["reason"] = "unknown"
        if not log_message:
            log_message = f'BackendError {status_code}: {data["reason"]}'

        super().__init__(log_message)
        self.status_code = status_code
        self.data = data


class UnknownError(BackendError):
    def __init__(self, data: Dict[str, Any], log_message: Optional[str] = None):
        super().__init__(500, data, log_message)


class BadRequest(BackendError):
    def __init__(self, data: Dict[str, Any], log_message: Optional[str] = None):
        if "reason" not in data:
            data["reason"] = "bad_request"
        super().__init__(400, data, log_message)


class UnauthorizedError(BackendError):
    def __init__(self, data: Dict[str, Any], log_message: Optional[str] = None):
        super().__init__(403, data, log_message)


class NotFoundError(BackendError):
    def __init__(self, data: Dict[str, Any], log_message: Optional[str] = None):
        super().__init__(404, data, log_message)



import datetime
import json
import os
from dataclasses import dataclass
from typing import Dict, Optional, Tuple, Type

import jwt
from jupyter_server.base.handlers import APIHandler
from jupyter_server.utils import url_path_join
from tornado.httputil import HTTPServerRequest
from tornado.routing import PathMatches, Rule, _RuleList

from aext_shared.backend_proxy import ProxyResponse, backend_proxy
from aext_shared.config import get_config
from aext_shared.errors import BackendError, UnauthorizedError

# TODO: This can be removed if/when we add anaconda-cloud-auth as a required dependency
try:
    from anaconda_cloud_auth.config import AuthConfig
    from anaconda_cloud_auth.exceptions import TokenNotFoundError
    from anaconda_cloud_auth.token import TokenInfo
except ImportError:  # pragma: no cover
    TokenInfo = None
    AuthConfig = None
    TokenNotFoundError = None

config = get_config()
debug_refresh_token = os.environ.get("DEBUG_REFRESH_TOKEN", None)
extra_anaconda_headers: Dict[str, str] = json.loads(os.environ.get("ANACONDA_CLOUD_EXTRA_HEADERS", "{}"))


@dataclass
class UserAccessCredentials:
    access_token: str = None
    new_access_token: str = None
    api_key: str = None
    username: str = None


def _get_api_key() -> Optional[str]:
    if TokenInfo is None:  # pragma: no cover
        # anaconda-cloud-auth is not installed
        return None
    auth_config = AuthConfig()
    try:
        return TokenInfo.load(domain=auth_config.domain).api_key
    except TokenNotFoundError:
        # We need to log in
        return None


def _get_username(token: Optional[str] = None) -> Optional[str]:
    try:
        if not token:
            token = _get_api_key()
        user_info = jwt.decode(token, algorithms=["RS256"], options={"verify_signature": False})
        return user_info.get("sub")
    except (TokenNotFoundError, jwt.exceptions.DecodeError):
        return None


class BackendHandler(APIHandler):
    def write_error(self, status_code, **kwargs):
        (err_cls, err, traceback) = kwargs.get("exc_info", (None, None, None))
        if isinstance(err, BackendError):
            self.set_status(err.status_code)
            self.finish(err.data)
            return
        self.set_status(status_code)
        self.finish({"reason": "unknown"})

    @classmethod
    def _is_token_expired(cls, token: str) -> bool:
        dt = datetime.datetime.now(datetime.timezone.utc)
        utc_timestamp = dt.replace(tzinfo=datetime.timezone.utc).timestamp()
        try:
            decoded_token = jwt.decode(token, options={"verify_signature": False})
        except jwt.DecodeError:
            return False
        # checking if the token expired - if the key is not found assume it is expired by returning utc_timestamp
        return utc_timestamp >= decoded_token.get("exp", utc_timestamp)

    async def get_user_access_credentials(self, auth_optional: bool = False) -> UserAccessCredentials:
        """
        Retrieves the access token used to proxy requests to anaconda.cloud API.

        Returns:
            - access_token: The current access_token
            - new_access_token: Returned in case it is a new access_token
            - api_key: Returns an API key in case it is a local stored API key
        """
        access_token = _get_cookie(self.request, "access_token")
        refresh_token = _get_cookie(self.request, "refresh_token")
        api_key = None
        new_access_token = None
        username = None

        if not access_token and not refresh_token and not debug_refresh_token:
            # Only attempt to get the local API key if neither cookie is set
            access_token = api_key = _get_api_key()

        if api_key is None:
            if not access_token or self._is_token_expired(access_token):
                # No matter what, try to update the access token
                # If auth is not required, ignore any failures that happen because of it
                try:
                    new_access_token, expires_in = await get_access_token(self.request)
                    access_token = new_access_token
                except BackendError as e:  # pragma: no cover
                    if not auth_optional:
                        raise e

        username = _get_username(access_token or new_access_token or api_key)
        return UserAccessCredentials(
            access_token=access_token,
            new_access_token=new_access_token,
            api_key=api_key,
            username=username,
        )

    async def anaconda_cloud_proxy(self, path: str, **kwargs) -> ProxyResponse:
        """
        This function proxies a request to anaconda.cloud API

        path: endpoint that should be hit
        kwargs: can be used to pass some request specifics such as:
                * headers
                * json: sets content-type to json
                * method: http verb [get, post, put, patch, delete]
        """
        auth_optional = kwargs.pop("auth_optional", False)
        expires_in = None
        user_access_credentials = await self.get_user_access_credentials(auth_optional=auth_optional)
        headers: Dict[str, str] = dict(kwargs.get("headers", {}))

        headers.update(extra_anaconda_headers)
        kwargs["headers"] = headers

        if user_access_credentials.access_token:
            headers["Authorization"] = f"Bearer {user_access_credentials.access_token}"

        # Enforce that the bearer token only gets sent to anaconda.cloud/api by setting the full URL in the helper method
        url = f"{config['cloud']['url']}/{path}"

        try:
            resp = await backend_proxy(self.request, url, **dict(kwargs))
        except BackendError as e:
            if user_access_credentials.api_key is not None:  # pragma: no cover
                # If we are using an API key, then we don't want to do the retry after refresh
                raise e

            if auth_optional or e.data.get("remote_status_code") not in [401, 403]:
                # Some unknown error, not due to expired token
                raise e

            # Refresh the token and retry the request
            user_access_credentials.new_access_token, expires_in = await get_access_token(self.request)
            headers["Authorization"] = f"Bearer {user_access_credentials.new_access_token}"
            resp = await backend_proxy(self.request, url, **dict(kwargs))
        finally:
            if user_access_credentials.new_access_token:
                self.set_cookie(
                    "access_token",
                    user_access_credentials.new_access_token,
                    max_age=expires_in,
                    secure=True,
                    samesite="strict",
                )
        return resp


def create_rules(base_url: str, prefix: str, handlers: Dict[str, Type[BackendHandler]]) -> _RuleList:
    rules: _RuleList = [
        Rule(PathMatches(url_path_join(base_url, prefix, key)), value) for key, value in handlers.items()
    ]
    return rules


def _get_cookie(request: HTTPServerRequest, name: str) -> Optional[str]:
    cookie = request.cookies.get(name)
    return cookie.value if cookie else None


async def get_access_token(incoming_request: HTTPServerRequest) -> Tuple[str, int]:
    refresh_token = _get_cookie(incoming_request, "refresh_token")
    if not refresh_token:
        refresh_token = debug_refresh_token

    if not refresh_token:
        raise UnauthorizedError({"reason": "missing refresh_token"})

    payload = {
        "grant_type": "refresh_token",
        "refresh_token": refresh_token,
    }
    try:
        proxy_response = await backend_proxy(
            incoming_request,
            f"{config['cloud']['url']}/iam/token",
            method="POST",
            json=payload,
            headers=dict(extra_anaconda_headers),
        )
    except BackendError as e:
        raise UnauthorizedError(e.data) from e
    access_token: str = proxy_response["remote_data"]["access_token"]
    expires_in: int = proxy_response["remote_data"]["expires_in"]
    return access_token, expires_in



import os
from typing import Any, Tuple, TypedDict

import httpx
from tornado.httputil import HTTPServerRequest

from aext_shared.config import get_config
from aext_shared.errors import BackendError, BadRequest, UnknownError

config = get_config()
cloudflare_client_id = os.environ.get("NUCLEUS_CLOUDFLARE_CLIENT_ID", None)
cloudflare_client_secret = os.environ.get("NUCLEUS_CLOUDFLARE_CLIENT_SECRET", None)


class ProxyResponse(TypedDict):
    remote_status_code: int
    remote_data: Any


async def backend_proxy(incoming_request: HTTPServerRequest, url: str, **kwargs) -> ProxyResponse:
    return (await backend_proxy_core(incoming_request=incoming_request, url=url, **kwargs))[0]


async def backend_proxy_core(
    incoming_request: HTTPServerRequest, url: str, **kwargs
) -> Tuple[ProxyResponse, httpx.Response]:
    headers: dict[str, str] = dict(kwargs.get("headers", {}))
    kwargs["headers"] = headers
    parse_response = kwargs.pop("parse_response", True)
    # Only add CF headers if the request is being triggered to Anaconda Cloud
    if f"{config['cloud']['url']}" in url:
        # Add (overridable) Cloudflare headers to internal requests
        if cloudflare_client_id and cloudflare_client_secret:
            headers.update(
                {"CF-Access-Client-Id": cloudflare_client_id, "CF-Access-Client-Secret": cloudflare_client_secret}
            )

    if "json" in kwargs:
        headers["Content-Type"] = "application/json"

    shared_headers = [(name, incoming_request.headers.get(name, None)) for name in ["User-Agent", "Api-Version"]]
    shared_headers = [(name, val) for (name, val) in shared_headers if val and name not in headers]
    for name, val in shared_headers:
        headers[name] = val

    method = kwargs.pop("method", "GET")
    proxy_request = httpx.Request(method, url, **kwargs)

    # Redirects strip Authorization headers if not same URL (cookies are smart and follow browser logic)
    # Attention should be paid to Cloudflare headers as they are not stripped (but are only set on staging for now).
    # More info:
    # https://github.com/encode/httpx/blob/053bc57c3799801ff11273dd393cb0715e63ecf9/httpx/_client.py#L459
    # https://github.com/encode/httpx/blob/053bc57c3799801ff11273dd393cb0715e63ecf9/httpx/_client.py#L536
    async with httpx.AsyncClient(follow_redirects=True) as client:
        try:
            response: httpx.Response = await client.send(proxy_request)
        except Exception as e:
            raise UnknownError({"reason": "proxy_connection_error"}) from e

        try:
            if not parse_response:
                return response, None
            data = response.json() if len(response.text) else None
        except ValueError as e:
            raise BadRequest(
                {
                    "reason": "proxy_response_not_json",
                    "remote_status_code": response.status_code,
                    "raw": response.text,
                }
            ) from e

        if response.status_code < httpx.codes.OK or response.status_code >= httpx.codes.MULTIPLE_CHOICES:
            raise BackendError(
                200,
                {
                    "reason": "proxy_bad_response",
                    "remote_status_code": response.status_code,
                    "remote_data": data,
                },
            )

        return ProxyResponse(remote_status_code=response.status_code, remote_data=data), response



import os
from unittest.mock import patch

import jupyterlab
import jupyterlab_server
import pytest

from aext_shared.config import get_config


@pytest.fixture
def default_config():
    return {
        "heap": {"clientId": "758475466"},
        "datadog": {"clientToken": "pub2b6467ca1f0ee597585d8a8bf43d523f", "service": "nb.anaconda.cloud"},
        "featureFlag": {
            "clientId": "6408b2b7e550011343183538",
            "overrides": {},
        },
        "cloud": {
            "url": "https://anaconda.cloud/api",
            "staticContent": "https://static.anaconda.cloud/shared",
        },
        "org": {
            "url": "https://notebooks.anaconda.org",
        },
        "notebooks": {"url": "https://nb.anaconda.cloud"},
        "environment": "development",
        "jupyterlab": {"server": jupyterlab_server.__version__, "client": jupyterlab.__version__},
        "assistant": {
            "environmentType": "cloud-notebooks-prod",
            "apiUrl": "https://assistant.anaconda.cloud",
        },
    }


@patch.dict(os.environ, {}, clear=True)
def test_get_default_config(default_config):
    assert get_config() == default_config


@patch.dict(
    os.environ,
    {
        "PYTHONANYWHERE_SITE": "https://nb.anaconda.cloud",
        "ANACONDA_NOTEBOOKS_URL": "https://nb.anaconda.cloud",
    },
    clear=True,
)
def test_get_config_with_production_environment(default_config):
    config = get_config()
    assert config["environment"] == "production"


@patch.dict(
    os.environ,
    {"DEBUG_AEXT_SHARED_LOCAL": "true"},
    clear=True,
)
def test_local_config(default_config):
    config = get_config()
    expected = default_config
    expected["heap"]["clientId"] = None
    expected["datadog"]["clientToken"] = None
    expected["datadog"]["service"] = None
    expected["featureFlag"]["clientId"] = None
    expected["featureFlag"]["overrides"] = {
        "monorepo-assistant": True,
        "monorepo-panel": True,
        "panel-sign-out-jupyterlab": True,
        "share-notebook-use-projects-v2": True,
        "sharing-show-preview": True,
        "monorepo-share-notebooks": True,
        "monorepo-panel-show-publish-button": False,
    }
    expected["environment"] = "local"
    expected["assistant"] = {
        "environmentType": "local-notebooks-prod",
        "apiUrl": "https://assistant.anaconda.cloud",
    }
    assert config == expected
